/*
 * Copyright 2020 Kunbus GmbH.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include<jni.h>
#include<piControlIf.hpp>
#include<piControl.h>
#include<string.h>

#include"com_kunbus_revpi_PiControl.h"


piControl *piC = NULL;

JNIEXPORT void JNICALL Java_com_kunbus_revpi_PiControl_init(JNIEnv *, jobject) {
    if (piC == NULL) {
        piC = new piControl();
    }
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_reset(JNIEnv *env, jobject thisObject) {
    return (piC->Reset());
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_read(JNIEnv *env, jobject thisObject, jlong offset, jlong length, jobject data) {
    unsigned int iData;

    int ret = piC->Read(offset, length, (uint8_t*) & iData);

    if (ret < 0)
        return (ret);

    jclass dataClass = env->GetObjectClass(data);
    jmethodID valueOfMethodID = env->GetMethodID(dataClass, "valueOf", "");

    env->CallLongMethod(data, valueOfMethodID, iData);

    return (ret);
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_write(JNIEnv *env, jobject thisObject, jlong offset, jlong length, jlong data) {

    unsigned int iData = data;
    return (piC->Write(offset, length, (uint8_t*) & iData));
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_getDeviceInfoList(JNIEnv *env, jobject thisObject, jobjectArray dev) {
    SDeviceInfoStr devLoc[255];
    int i;
    int ret;

    if ((ret = piC->GetDeviceInfoList(devLoc)) < 0)
        return (ret);

    jsize devArrayLength = env->GetArrayLength(dev);
    if (devArrayLength < ret)
        return (-11);

    jobject devEntry = env->GetObjectArrayElement(dev, 0);
    jclass sDeviceInfoStrClass = env->GetObjectClass(devEntry);

    jfieldID addressFieldID = env->GetFieldID(sDeviceInfoStrClass, "address", "I");
    jfieldID serialnumberFieldID = env->GetFieldID(sDeviceInfoStrClass, "serialnumber", "J");
    jfieldID moduleTypeFieldID = env->GetFieldID(sDeviceInfoStrClass, "moduleType", "I");
    jfieldID hwRevisionFieldID = env->GetFieldID(sDeviceInfoStrClass, "hwRevision", "I");
    jfieldID swMajorFieldID = env->GetFieldID(sDeviceInfoStrClass, "swMajor", "I");
    jfieldID swMinorFieldID = env->GetFieldID(sDeviceInfoStrClass, "swMinor", "I");
    jfieldID svnRevisionFieldID = env->GetFieldID(sDeviceInfoStrClass, "svnRevision", "J");
    jfieldID inputLengthFieldID = env->GetFieldID(sDeviceInfoStrClass, "inputLength", "I");
    jfieldID outputLengthFieldID = env->GetFieldID(sDeviceInfoStrClass, "outputLength", "I");
    jfieldID configLengthFieldID = env->GetFieldID(sDeviceInfoStrClass, "configLength", "I");
    jfieldID baseOffsetFieldID = env->GetFieldID(sDeviceInfoStrClass, "baseOffset", "I");
    jfieldID inputOffsetFieldID = env->GetFieldID(sDeviceInfoStrClass, "inputOffset", "I");
    jfieldID outputOffsetFieldID = env->GetFieldID(sDeviceInfoStrClass, "outputOffset", "I");
    jfieldID configOffsetFieldID = env->GetFieldID(sDeviceInfoStrClass, "configOffset", "I");
    jfieldID firstEntryFieldID = env->GetFieldID(sDeviceInfoStrClass, "firstEntry", "I");
    jfieldID entriesFieldID = env->GetFieldID(sDeviceInfoStrClass, "entries", "I");
    jfieldID moduleStateFieldID = env->GetFieldID(sDeviceInfoStrClass, "moduleState", "I");
    jfieldID activeFieldID = env->GetFieldID(sDeviceInfoStrClass, "active", "I");

    for (i = 0; i < ret; i++) {
        jobject devEntry = env->GetObjectArrayElement(dev, i);

        if (addressFieldID == NULL || serialnumberFieldID == NULL || moduleTypeFieldID == NULL ||
                hwRevisionFieldID == NULL || swMajorFieldID == NULL || swMinorFieldID == NULL || svnRevisionFieldID == NULL
                || inputLengthFieldID == NULL || outputLengthFieldID == NULL || configLengthFieldID == NULL ||
                baseOffsetFieldID == NULL || inputOffsetFieldID == NULL || outputOffsetFieldID == NULL ||
                configOffsetFieldID == NULL || firstEntryFieldID == NULL || entriesFieldID == NULL ||
                moduleStateFieldID == NULL || activeFieldID == NULL)
            return (-12);

        env->SetIntField(sDeviceInfoStrClass, addressFieldID, devLoc[i].i8uAddress);
        env->SetIntField(sDeviceInfoStrClass, serialnumberFieldID, devLoc[i].i32uSerialnumber);
        env->SetIntField(sDeviceInfoStrClass, moduleTypeFieldID, devLoc[i].i16uModuleType);
        env->SetIntField(sDeviceInfoStrClass, hwRevisionFieldID, devLoc[i].i16uHW_Revision);
        env->SetIntField(sDeviceInfoStrClass, swMajorFieldID, devLoc[i].i16uSW_Major);
        env->SetIntField(sDeviceInfoStrClass, swMinorFieldID, devLoc[i].i16uSW_Minor);
        env->SetIntField(sDeviceInfoStrClass, svnRevisionFieldID, devLoc[i].i32uSVN_Revision);
        env->SetIntField(sDeviceInfoStrClass, inputLengthFieldID, devLoc[i].i16uInputLength);
        env->SetIntField(sDeviceInfoStrClass, outputLengthFieldID, devLoc[i].i16uOutputLength);
        env->SetIntField(sDeviceInfoStrClass, configLengthFieldID, devLoc[i].i16uConfigLength);
        env->SetIntField(sDeviceInfoStrClass, baseOffsetFieldID, devLoc[i].i16uBaseOffset);
        env->SetIntField(sDeviceInfoStrClass, inputOffsetFieldID, devLoc[i].i16uInputOffset);
        env->SetIntField(sDeviceInfoStrClass, outputOffsetFieldID, devLoc[i].i16uOutputOffset);
        env->SetIntField(sDeviceInfoStrClass, configOffsetFieldID, devLoc[i].i16uConfigOffset);
        env->SetIntField(sDeviceInfoStrClass, firstEntryFieldID, devLoc[i].i16uFirstEntry);
        env->SetIntField(sDeviceInfoStrClass, entriesFieldID, devLoc[i].i16uEntries);
        env->SetIntField(sDeviceInfoStrClass, moduleStateFieldID, devLoc[i].i8uModuleState);
        env->SetIntField(sDeviceInfoStrClass, activeFieldID, devLoc[i].i8uActive);
    }

    return (ret);
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_getBitValue(JNIEnv *env, jobject thisObject, jobject spiValue) {
    SPIValue spiValueLoc;

    jclass spiValueClass = env->GetObjectClass(spiValue);

    jfieldID addressFieldID = env->GetFieldID(spiValueClass, "address", "I");
    jfieldID bitFieldID = env->GetFieldID(spiValueClass, "bit", "I");
    jfieldID valueFieldID = env->GetFieldID(spiValueClass, "value", "I");

    if (addressFieldID == NULL || bitFieldID == NULL || valueFieldID == NULL)
        return (-11);

    spiValueLoc.i16uAddress = env->GetIntField(spiValue, addressFieldID);
    spiValueLoc.i8uBit = env->GetIntField(spiValue, bitFieldID);
    spiValueLoc.i8uValue = env->GetIntField(spiValue, valueFieldID);

    int ret = piC->SetBitValue(&spiValueLoc);

    env->SetIntField(spiValue, addressFieldID, spiValueLoc.i16uAddress);
    env->SetIntField(spiValue, bitFieldID, spiValueLoc.i8uBit);
    env->SetIntField(spiValue, valueFieldID, spiValueLoc.i8uValue);

    return (ret);
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_setBitValue(JNIEnv *env, jobject thisObject, jobject spiValue) {
    SPIValue spiValueLoc;

    jclass spiValueClass = env->GetObjectClass(spiValue);

    jfieldID addressFieldID = env->GetFieldID(spiValueClass, "address", "I");
    jfieldID bitFieldID = env->GetFieldID(spiValueClass, "bit", "I");
    jfieldID valueFieldID = env->GetFieldID(spiValueClass, "value", "I");

    if (addressFieldID == NULL || bitFieldID == NULL || valueFieldID == NULL)
        return (-11);

    spiValueLoc.i16uAddress = env->GetIntField(spiValue, addressFieldID);
    spiValueLoc.i8uBit = env->GetIntField(spiValue, bitFieldID);
    spiValueLoc.i8uValue = env->GetIntField(spiValue, valueFieldID);

    int ret = piC->SetBitValue(&spiValueLoc);

    env->SetIntField(spiValue, addressFieldID, spiValueLoc.i16uAddress);
    env->SetIntField(spiValue, bitFieldID, spiValueLoc.i8uBit);
    env->SetIntField(spiValue, valueFieldID, spiValueLoc.i8uValue);

    return (ret);

}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_getVariableInfo(JNIEnv *env, jobject thisObject, jobject spiVariable) {
    SPIVariable spiVariableLoc;

    jclass spiVariableClass = env->GetObjectClass(spiVariable);

    jfieldID varNameFieldID = env->GetFieldID(spiVariableClass, "varName", "Ljava/lang/String;");
    jfieldID addressFieldID = env->GetFieldID(spiVariableClass, "address", "I");
    jfieldID bitFieldID = env->GetFieldID(spiVariableClass, "bit", "I");
    jfieldID lengthFieldID = env->GetFieldID(spiVariableClass, "length", "I");

    if (varNameFieldID == NULL || addressFieldID == NULL || bitFieldID == NULL || lengthFieldID == NULL)
        return (-11);

    jstring varName = (jstring) env->GetObjectField(spiVariable, varNameFieldID);
    const char *varNameCStr = env->GetStringUTFChars(varName, NULL);
    if (varNameCStr == NULL)
        return (-12);

    jsize len = env->GetStringLength(varName);
    bool lenOk = len < sizeof (SPIVariable::strVarName);

    if (lenOk) {
        strcpy(spiVariableLoc.strVarName, varNameCStr);
        spiVariableLoc.strVarName[31] = 0;
    }

    env->ReleaseStringUTFChars(varName, varNameCStr);

    if (!lenOk) {
        return (-13);
    }

    spiVariableLoc.i16uAddress = env->GetIntField(spiVariable, addressFieldID);
    spiVariableLoc.i8uBit = env->GetIntField(spiVariable, bitFieldID);
    spiVariableLoc.i16uLength = env->GetIntField(spiVariable, lengthFieldID);

    int ret = piC->GetVariableInfo(&spiVariableLoc);

    env->SetIntField(spiVariable, addressFieldID, spiVariableLoc.i16uAddress);
    env->SetIntField(spiVariable, bitFieldID, spiVariableLoc.i8uBit);
    env->SetIntField(spiVariable, lengthFieldID, spiVariableLoc.i16uLength);

    return (ret);
}

JNIEXPORT jint JNICALL Java_com_kunbus_revpi_PiControl_findVariable(JNIEnv *env, jobject thisObject, jstring name) {

    const char *varNameCStr = env->GetStringUTFChars(name, NULL);
    if (varNameCStr == NULL)
        return (-1);

    int ret = piC->FindVariable(varNameCStr);

    env->ReleaseStringUTFChars(name, varNameCStr);

    return (ret);
}